import std/[json, options, os, paths, sequtils, strutils, terminal, times,
            unicode]
import cliutils, docopt, namespaced_logging, mpv

import wdiwtlt/[cliconstants, commandbuffer, logging, medialibrary,
                messagebuffer, models, nonblockingio, scrolltext, usage]

type
  CliMode {.pure.} = enum Direct, Command

  MessageBuffers* = ref object
    playing: MessageBuffer[ScrollText]
    status: MessageBuffer[ScrollText]
    longform: MessageBuffer[string]
    input: MessageBuffer[string]

  History = ref object
    entries: seq[string]
    idx: int

  CliContext = ref object
    cfg: CombinedConfig
    cmd: CommandBuffer
    curMediaFile: Option[MediaFile]
    lib: MediaLibrary
    log: Option[Logger]
    logs: seq[LogMessage]
    logThreshold: Level
    mode: CliMode
    msgs: MessageBuffers
    mpv: ptr handle
    msgDuration: Duration
    stop*: bool
    width*: int
    height*: int

var ctx {.threadvar.}: CliContext

const FRAME_DUR_MS = 200

proc initMpv(): ptr handle =
  result = mpv.create()

  if result.isNil:
    raise newException(ValueError, "failed creating mpv context")

  # Enable default key bindings, so the user can actually interact with
  # the player (and e.g. close the window).
  result.set_option("terminal", "no")
  result.set_option("input-default-bindings", "yes")
  result.set_option("input-vo-keyboard", "yes")
  result.set_option("osc", false)
  #result.set_property("volume", "50.0")
  discard result.request_log_messages("no")

  # Done setting up options.
  check_error result.initialize()


proc cleanup() =
  terminate_destroy(ctx.mpv)
  setNonBlockingTty(false)
  echo ""


#[
proc exitHandlerHook() {.noconv.} =
  # This is called when the user presses Ctrl+C
  # or when the program exits normally.
  terminate_destroy(ctx.mpv)
  echo ""
  quit(QuitSuccess)
]#

proc initContext(args: Table[string, Value]): CliContext =
  var wdiwtltCfgFilename: string

  let log = getLogger("wdiwtlt")

  let cfgLocations =
    if args["--config"]: @[$args["--config"]]
    elif existsEnv("XDG_CONFIG_HOME"): @[getEnv("XDG_CONFIG_HOME") / "wdiwtlt.json"]
    else: @[$getEnv("HOME") / ".config" / "wdiwtlt.json" ]

  try: wdiwtltCfgFilename = findConfigFile(".wdiwtlt.json", cfgLocations)
  except ValueError:
    log.error "could not find wdiwtlt config file: " & wdiwtltCfgFilename
    if not args["--config"]:
      wdiwtltCfgFilename = cfgLocations[0]

      var cfgFile: File
      try:
        cfgFile = open(wdiwtltCfgFilename, fmWrite)
        cfgFile.write((%*{
          "libraryRoot":
            if args["--library-root"]:
              $args["--library-root"]
            else:
              getEnv("XDG_MUSIC_DIR", getEnv("HOME") / "music"),
          "dbPath": getEnv("XDG_DATA_HOME", (getEnv("HOME") / ".local/share")) / "wdiwtlt/db.sqlite",
        }).pretty)
        log.info "created sample config file at " & wdiwtltCfgFilename
      except CatchableError:
        log.error "could not write default .wdiwtlt.json to " &
          wdiwtltCfgFilename
      finally: close(cfgFile)

    quit(QuitFailure)

  log.debug("loading config from '$#'" % [wdiwtltCfgFilename])
  let cfg = initCombinedConfig(wdiwtltCfgFilename, args)

  let width = min(terminalWidth(), cfg.getVal("maxWidth", "80").parseInt)
  let height = min(terminalHeight(), cfg.getVal("maxHeight", "60").parseInt)

  let msgs = MessageBuffers(
    playing: initMessageBuffer[ScrollText](
      dims = BufferDimensions(width: width, height: 1, x: 0, y: 0),
      initialMsg = "Nothing playing"),
    status: initMessageBuffer[ScrollText](
      dims = BufferDimensions(width: width, height: 1, x: 0, y: 1),
      initialMsg = ""),
    input: initMessageBuffer[string](
      dims = BufferDimensions(width: width, height: 1, x: 0, y: 2),
      initialMsg = ""),
    longform: initMessageBuffer[string](
      dims = BufferDimensions(width: width, height: height - 3, x: 0, y: 4),
      initialMsg = ""))

  result = CliContext(
    cfg: cfg,
    cmd: initCommandBuffer(msgs.input),
    curMediaFile: none(MediaFile),
    lib: initMediaLibrary(
      rootDir = getVal(cfg, "libraryRoot").Path,
      dbPath = getVal(cfg, "dbPath").Path),
    log: log,
    mode: CliMode.Direct,
    msgs: msgs,
    mpv: initMpv(),
    msgDuration: initDuration(seconds = 5, milliseconds = 0),
    stop: false,
    width: width,
    height: height)

  if logService.isSome:
    let customLogAppender = initCustomLogAppender(doLogMessage =
      proc (msg: LogMessage): void = ctx.logs.add(msg))

    logService.get.clearAppenders()
    logService.get.addAppender(customLogAppender)


proc nextTick(ctx: var CliContext) =
  ctx.msgs.playing.nextTick
  ctx.msgs.status.nextTick
  ctx.msgs.longform.nextTick


proc setMode(ctx: var CliContext, mode: CliMode) =
  ctx.mode = mode
  case mode

  of CliMode.Direct:
    hideCursor()

  of CliMode.Command:
    showCursor()
    ctx.cmd.setMode(EditMode.Insert)


proc statusMsg(ctx: var CliContext, msg: string) =
  ctx.msgs.status.showMsg(msg, some(ctx.msgDuration))


proc longMsg[T: string or seq[string]](ctx: var CliContext, msg: T) =
  ctx.msgs.longform.showMsg(msg, some(ctx.msgDuration))


proc viewLogLoops(ctx: var CliContext) =
  let logLines = ctx.logs
    .filterIt(it.level > ctx.logThreshold)
    .mapIt("$# - [$#]: $#" % [$it.level, $it.scope, it.message])
    .join("\p")

  ctx.msgs.longform.showMsg(logLines, none[Duration]())

  var viewLogs = true
  while viewLogs:
    case getKeyAsync()
    of Key.Q: viewLogs = false
    of Key.K, Key.Up: ctx.msgs.longform.scrollLines(-1)
    of Key.J, Key.Down: ctx.msgs.longform.scrollLines(1)
    of Key.H, Key.Left: ctx.msgs.longform.scrollColumns(-1)
    of Key.L, Key.Right: ctx.msgs.longform.scrollColumns(1)
    of Key.PageUp: ctx.msgs.longform.scrollPages(-1)
    of Key.PageDown: ctx.msgs.longform.scrollPages(1)
    else: discard

    sleep(FRAME_DUR_MS)

  ctx.msgs.longform.clear()

proc scanLibrary(ctx: var CliContext) =
  ctx.statusMsg("Scanning media library...")
  let counts = ctx.lib.rescanLibrary()

  var scanResults = @[
    "─".repeat(ctx.width),
    "Scan complete:",
    "\t$# files total." % [ $counts.total ]
  ]

  if counts.new > 0:
    scanResults.add("\t$# new files added." % [ $counts.new ])

  if counts.ignored > 0:
    scanResults.add("\t$# files ignored." % [ $counts.ignored ])

  if counts.absent > 0:
    scanResults.add(
      "\t$# files in the database but not stored locally." %
      [ $counts.absent ])

  ctx.longMsg(scanResults.join("\p"))

proc processLogs(ctx: var CliContext, args: seq[string]) =
  case args[0]
  of "view": ctx.viewLogLoops()
  of "clear": ctx.logs = @[]
  of "set-threshold":
    if args.len < 2:
      ctx.longMsg(("missing $# argument to $#" %
        [ "log-threshold", "logs set-threshold" ]) &
        usageOf(@["logs", "set-threshold"]))

    # TODO
  else: ctx.statusMsg("Unrecognized logs command: " & args[0])


proc processCommand(ctx: var CliContext, command: string) =
  let parts = command.strip.split(' ')

  case parts[0]
  of "scan":                scanLibrary(ctx)
  of "q", "quit", "exit":   ctx.stop = true
  of "logs":                ctx.processLogs(parts[1..^1])
  else: statusMsg(ctx, "Unrecognized command: " & command)


proc handleKey(ctx: var CliContext, key: Key) =
  if ctx.mode == CliMode.Direct:
    case key
    of Key.Q:     ctx.stop = true
    of Key.Colon, Key.I: ctx.setMode(CliMode.Command)
    else: discard

  elif ctx.mode == CliMode.Command:
    case key
    of Key.Enter:
      let command = $ctx.cmd
      ctx.cmd.clear
      ctx.cmd.mode = EditMode.Insert
      processCommand(ctx, command)
    of Key.Backspace: ctx.cmd.handleInput(Key.Backspace)
    of Key.Escape:
      if ctx.cmd.mode == EditMode.Command:
        ctx.setMode(CliMode.Direct)
        ctx.cmd.clear
      else: ctx.cmd.handleInput(Key.Escape)
    else: ctx.cmd.handleInput(key)


proc mainLoop(ctx: var CliContext) =

  hideCursor()
  var frame = 0

  while not ctx.stop:

    # This sleep below will be the primary driver of the frame rate of the
    # application (examples below):
    #
    # | frames/sec | sec/frame | ms/frame |
    # |------------|-----------|----------|
    # |    5.00    |   0.200   |    200   |
    # |   10.00    |   0.100   |    100   |
    # |   12.50    |   0.080   |     80   |
    # |   20.00    |   0.050   |     50   |
    # |   25.00    |   0.040   |     40   |
    # |   31.25    |   0.032   |     32   |
    # |   50.00    |   0.020   |     20   |
    # |   62.50    |   0.016   |     16   |
    #
    # Previously, when we were using illwill and rendering every frame we
    # targeted a faster FPS to allow for a responsive feeling to the command
    # input, etc. In this case, we still had a "logic update" period that was
    # lower than our rendering FPS. Specifically, we were targeting 60fps
    # rendering speed, but 5fps for logic updates (scrolling text, expiring
    # messages, etc.).
    #
    # Now that we are directly rendering (components of the UI render updates
    # themselves immediately) we don't need to render at 60fps to have a
    # responsive UI. Now we really only care about the frequency of logic
    # updates, so we're targeting 5fps and triggering our logic updates every
    # frame.
    handleKey(ctx, getKeyAsync())
    sleep(FRAME_DUR_MS)
    nextTick(ctx) # tick on every frame
    frame = (frame + 1) mod 1000 # use a max frame count to prevent overflow


when isMainModule:

  discard enableLogging()
  try:
    let args = docopt(USAGE, version=VERSION)

    ctx = initContext(args)
    setNonBlockingTty(true)

    mainLoop(ctx)

  except Exception:
    let ex = getCurrentException()
    getLogger("wdiwtlt").error(%*{
      "msg": "Unhandled exception",
      "error": ex.msg,
      "trace": ex.getStackTrace()
    })
    cleanup()
    quit(QuitFailure)
  finally:
    cleanup()
