import std/[options, strutils]

import ./[messagebuffer, nonblockingio, terminal, util]

type
  EditMode* {.pure.} = enum Command, Insert, Overwrite, Visual

  Command = ref object
    buffer: string
    idx: int
    selectionStartIdx: Option[int]

  CommandBuffer* = ref object
    history: seq[Command]
    mb: MessageBuffer[string]
    idx: int
    visibleIdx: int # first visible character in the cmd buffer
    mode*: EditMode

func initCommandBuffer*(mb: MessageBuffer[string]): CommandBuffer =
  result = CommandBuffer(
    history: @[Command(
      buffer: "",
      idx: 0,
      selectionStartIdx: none(int))],
    idx: 0,
    mb: mb,
    mode: Insert)

proc cur(cb: CommandBuffer): Command = cb.history[cb.idx]
proc cur(cb: var CommandBuffer): Command = cb.history[cb.idx]

proc prev*(cb: var CommandBuffer) =
  cb.idx = clamp( cb.idx - 1, 0, cb.history.len - 1)

proc next*(cb: var CommandBuffer) =
  cb.idx = clamp( cb.idx + 1, 0, cb.history.len - 1)

proc insert*(cb: var CommandBuffer, s: string) =
  var cmd = cb.history[cb.idx]
  cmd.buffer = cmd.buffer[0..<cmd.idx] & s & cmd.buffer[cmd.idx..^1]
  cmd.idx += s.len

proc overwrite*(cb: var CommandBuffer, s: string) =
  var cmd = cb.history[cb.idx]
  if cmd.idx + s.len > cmd.buffer.len:
    cmd.buffer = cmd.buffer[0..<cmd.idx] & s
  else:
    cmd.buffer = cmd.buffer[0..<cmd.idx] & s & cmd.buffer[cmd.idx + s.len..^1]
  cmd.idx += s.len

proc write*(cb: var CommandBuffer, s: string) =
  if cb.mode == EditMode.Insert: cb.insert(s)
  elif cb.mode == EditMode.Overwrite: cb.overwrite(s)

proc delete*(cb: var CommandBuffer, backspace = true) =
  var cmd = cb.history[cb.idx]
  if backspace and cmd.idx > 0:
    cmd.buffer = cmd.buffer[0..<cmd.idx - 1] & cmd.buffer[cmd.idx..^1]
    cmd.idx -= 1
  elif not backspace and cmd.idx < cmd.buffer.len:
    cmd.buffer = cmd.buffer[0..<cmd.idx] & cmd.buffer[cmd.idx + 1..^1]

proc clear*(cb: var CommandBuffer) =
  cb.history.add(Command(
    buffer: "",
    idx: 0,
    selectionStartIdx: none(int)))
  cb.idx = cb.history.len - 1
  cb.visibleIdx = 0

proc left*(cb: var CommandBuffer) =
  cb.cur.idx = clamp(cb.cur.idx - 1, 0, cb.cur.buffer.len)

proc right*(cb: var CommandBuffer) =
  cb.cur.idx = clamp(cb.cur.idx + 1, 0, cb.cur.buffer.len)

proc toHome*(cb: var CommandBuffer) = cb.cur.idx = 0
proc toEnd*(cb: var CommandBuffer) = cb.cur.idx = cb.cur.buffer.len

proc backWord*(cb: var CommandBuffer) =
  var cmd = cb.history[cb.idx]
  while cmd.idx > 0 and cmd.buffer[cmd.idx - 1] == ' ': cmd.idx -= 1
  while cmd.idx > 0 and cmd.buffer[cmd.idx - 1] != ' ': cmd.idx -= 1

proc forwardWord*(cb: var CommandBuffer) =
  var cmd = cb.history[cb.idx]
  while cmd.idx < cmd.buffer.len and cmd.buffer[cmd.idx] != ' ': cmd.idx += 1
  while cmd.idx < cmd.buffer.len and cmd.buffer[cmd.idx] == ' ': cmd.idx += 1

proc endOfWord*(cb: var CommandBuffer) =
  var cmd = cb.history[cb.idx]
  cmd.idx = clamp(cmd.idx + 1, 0, cmd.buffer.len - 1)
  while cmd.idx < cmd.buffer.len and cmd.buffer[cmd.idx] == ' ': cmd.idx += 1
  while cmd.idx < cmd.buffer.len - 1 and
        cmd.buffer[cmd.idx + 1] != ' ':
    cmd.idx += 1

func toChar(k: Key): char = cast[char](ord(k))


proc setMode*(cb: var CommandBuffer, mode: EditMode) =
  cb.mode = mode
  case mode
  of EditMode.Insert:     setCursorType(ctBarBlink)
  of EditMode.Overwrite:  setCursorType(ctBlock)
  else:                   setCursorType(ctBlockBlink)


func `$`*(cb: CommandBuffer): string =
  if cb.cur.selectionStartIdx.isNone:
    return cb.cur.buffer
  else:
    let selIdx = cb.cur.selectionStartIdx.get
    let curIdx = cb.cur.idx

    if selIdx < curIdx:
      return cb.cur.buffer[0..<selIdx] &
        invert(cb.cur.buffer[selIdx..<curIdx]) &
        cb.cur.buffer[curIdx..^1]
    else:
      return cb.cur.buffer[0..<curIdx] &
        invert(cb.cur.buffer[curIdx..<selIdx]) &
        cb.cur.buffer[selIdx..^1]

func idx*(cb: CommandBuffer): int = cb.cur.idx


proc render*(cb: var CommandBuffer) =
  write(cb.mb.dims.x, cb.mb.dims.y, ' '.repeat(cb.mb.dims.width))

  let cmd = cb.cur
  var inputLine = cmd.buffer
  let spaceAvailable = cb.mb.dims.width - 2

  # Check to see that the line entered fits entirely within the space
  # available to display it. We need two characters of space: one for the
  # colon and one for the cursor. If we don't have enough space, we need to
  # decide which portion of the line to display.
  if len(cmd.buffer) >= spaceAvailable:
    # If the user has moved the cursor left of our visible index, follow them
    # back
    if cmd.idx < cb.visibleIdx: cb.visibleIdx = cmd.idx

    # If the user has moved right, push our index to the right to follow them
    elif (cmd.idx - cb.visibleIdx) > spaceAvailable:
      cb.visibleIdx = cmd.idx - spaceAvailable

    # If the user has started backspacing (we are showing less that we
    # could), follow them back
    elif (len(cmd.buffer) - cb.visibleIdx) < spaceAvailable:
      cb.visibleIdx = len(cmd.buffer) - spaceAvailable

    # Show the portion of the line starting at the visible index and ending
    # either with the end of the line or the end of the visible space
    # (whichever comes first). We need to check both because the user may
    # delete characters from the end of the line, which would make the
    # portion of the line we're currently showing shorter than the visible
    # space (don't try to access an index past the end of the string).
    inputLine = cmd.buffer[
      cb.visibleIdx ..<
      min(cb.visibleIdx + spaceAvailable, len(cmd.buffer))]

  elif cb.visibleIdx > 0:
    # We know the line fits within the space available, but we're still
    # showing a subset of the line (probably because it was longer and the
    # user backspaced). Let's just reset and show the whole line now that we
    # can.
    cb.visibleIdx = 0

  # TODO: implement VISUAL mode selection highlighting
  cb.mb.showMsg(":" & inputLine)
  setCursorPosition(
    cb.mb.dims.x + (cmd.idx - cb.visibleIdx) + 1,
    cb.mb.dims.y)

proc handleInput*(cb: var CommandBuffer, key: Key) =
  case cb.mode
  of EditMode.Insert, EditMode.Overwrite:
    case key
    of Key.Escape: cb.setMode(EditMode.Command)
    of Key.Backspace: cb.delete()
    of Key.Delete: cb.delete(backspace = false)
    of Key.Left: cb.left()
    of Key.Right: cb.right()
    of Key.Up: cb.prev()
    of Key.Down: cb.next()
    of Key.Home: cb.toHome()
    of Key.End: cb.toEnd()
    of Key.CtrlH: cb.backWord()
    of Key.CtrlL: cb.forwardWord()
    elif key >= Key.Space and key <= Key.Tilde: cb.write($toChar(key))
    else: discard

  of EditMode.Command:
    case key
    of Key.Backspace: cb.delete()
    of Key.Delete,  Key.X: cb.delete(backspace = false)
    of Key.Left,    Key.H: cb.left()
    of Key.Right,   Key.L: cb.right()
    of Key.Up,      Key.J: cb.prev()
    of Key.Down,    Key.K: cb.next()
    of Key.Home,    Key.Zero:   cb.toHome()
    of Key.End,     Key.Dollar: cb.toEnd()
    of Key.B: cb.backWord()
    of Key.W: cb.forwardWord()
    of Key.E: cb.endOfWord()
    of Key.I: cb.setMode(EditMode.Insert)
    of Key.A:
      cb.cur.idx = clamp(cb.cur.idx + 1, 0, cb.cur.buffer.len)
      cb.setMode(EditMode.Insert)
    of Key.ShiftA:
      cb.cur.idx = cb.cur.buffer.len
      cb.setMode(EditMode.Insert)
    of Key.ShiftI:
      cb.cur.idx = 0
      cb.setMode(EditMode.Insert)
    of Key.ShiftR:
      cb.setMode(EditMode.Overwrite)
    of Key.V:
      cb.setMode(EditMode.Visual)
      cb.cur.selectionStartIdx = some(cb.cur.idx)
    else: discard

  of EditMode.Visual:
    case key
    of Key.Escape:
      cb.setMode(EditMode.Command)
      cb.cur.selectionStartIdx = none(int)
    of Key.Backspace: cb.left()
    of Key.Left,   Key.H: cb.left()
    of Key.Right,  Key.L: cb.right()
    of Key.Up,     Key.J:
      cb.cur.selectionStartIdx = none[int]()
      cb.prev()
    of Key.Down,   Key.K:
      cb.cur.selectionStartIdx = none[int]()
    of Key.Home,   Key.Zero: cb.toHome()
    of Key.End,    Key.Dollar: cb.toEnd()
    of Key.B: cb.backWord()
    of Key.W: cb.forwardWord()
    of Key.V:
      cb.setMode(EditMode.Command)
      cb.cur.selectionStartIdx = none(int)
    else: discard

  render(cb)
