import std/[options, sequtils, strutils, times, wordwrap]

import ./[scrolltext, terminal, util]

type
  BufferDimensions* = object
    width*, height*, x*, y*: int

  MessageBuffer*[T: ScrollText or string] = ref object
    content: T
    dims*: BufferDimensions
    dismissAfter: Option[DateTime]
    firstColumnIdx: int
    firstLineIdx: int


proc getLinePortion(line: string, start: int, width: int): string =
  let cleanLine = stripFormatting(line)

  if start == 0 and len(cleanLine) <= width:
      result = line & ' '.repeat(width - len(cleanLine))
  else:
    result = ansiAwareSubstring(line, start, width)


proc render[T: ScrollText or string](mb: var MessageBuffer[T]) =
  hideCursor()
  when T is ScrollText: write(mb.dims.x, mb.dims.y, $mb.content)
  when T is string:
    let renderedContent = splitLines(mb.content)
      .mapIt(wrapWords(
        s = it,
        maxLineWidth = mb.dims.width,
        newLine = "\p").alignLeft(mb.dims.width))
      .mapIt(getLinePortion(it, mb.firstColumnIdx, mb.dims.width))
      .join("\p")
    let lastLineIdx = min(mb.firstLineIdx + mb.dims.height, len(renderedContent))
    write(mb.dims.x, mb.dims.y, renderedContent[mb.firstLineIdx..<lastLineIdx])
  showCursor()


func variableMsgDuration[T: ScrollText or string](
    mb: MessageBuffer[T],
    initDur: Duration): Duration =
  when T is ScrollText:
    # text moves at 5 characters per second (192ms per character)
    result = initDur + initDuration(milliseconds = len($mb.content) * 192)
  when T is string:
    result = initDur + initDuration(milliseconds = len(splitLines(mb.content)) * 250)


proc initMessageBuffer*[T: ScrollText or string](
    dims: BufferDimensions,
    initialMsg = ""): MessageBuffer[T] =

  when T is ScrollText:
    result = MessageBuffer[T](
      content: initScrollText(initialMsg, max(dims.width, 1)),
      dims: BufferDimensions(
        width: max(dims.width, 1),
        height: 1,
        x: dims.x,
        y: dims.y),
      dismissAfter: none[DateTime](),
      firstColumnIdx: 0,
      firstLineIdx: 0)

  when T is string:
    result = MessageBuffer[T](
      content: initialMsg,
      dims: BufferDimensions(
        width: max(dims.width, 1),
        height: dims.height,
        x: dims.x,
        y: dims.y),
      dismissAfter: none[DateTime](),
      firstColumnIdx: 0,
      firstLineIdx: 0)

  render(result)


proc showMsg*[T: ScrollText or string](
    mb: var MessageBuffer[T],
    msg: string,
    dur: Option[Duration] = none[Duration]()) =

  when T is ScrollText: mb.content.text = msg
  when T is string: mb.content = msg

  if dur.isSome:
    mb.dismissAfter = some(now() + mb.variableMsgDuration(dur.get))
  else:
    mb.dismissAfter = none[DateTime]()

  render(mb)


proc clear*[T: ScrollText or string](mb: var MessageBuffer[T]) =
  when T is ScrollText:
    let clearText = ' '.repeat(mb.dims.width)
    mb.content.text = ""

  when T is string:
    let clearText = splitLines(mb.content)
      .mapIt(' '.repeat(mb.dims.width))
      .join("\p")

    mb.content = ""

  hideCursor()
  write(mb.dims.x, mb.dims.y, clearText)
  showCursor()
  mb.dismissAfter = none[DateTime]()


#[ TODO: re-implement taking BufferDimensions if needed
proc `maxWidth=`*[T: ScrollText or string](
    mb: var MessageBuffer[T],
    maxWidth: int) =
  mb.maxWidth = max(maxWidth, 1)

  when T is ScrollText: mb.content.maxWidth = mb.maxWidth
  when T is string: mb.renderStringContent()
]#


proc nextTick*[T: ScrollText or string](mb: var MessageBuffer[T]) =

  if mb.dismissAfter.isSome and now() > mb.dismissAfter.get:
    mb.clear()

  when T is ScrollText:
    if mb.content.isScrolling:
      discard mb.content.nextTick()
      render(mb)


func hasContent*[T: ScrollText or string](
    mb: MessageBuffer[T]): bool =
  when T is ScrollText: return len($mb.content) > 0
  when T is string: return len(mb.content) > 0


proc scrollLines*[T](mb: var MessageBuffer[T], lines: int) =
  mb.firstLineIdx = clamp(
    mb.firstLineIdx + lines,
    0,
    max(0, len(mb.content) - mb.dims.height))
  render(mb)

proc scrollPages*[T](mb: var MessageBuffer[T], pages: int) =
  scrollLines(mb, pages * mb.dims.height)


proc scrollColumns*(mb: var MessageBuffer[string], chars: int) =
  mb.firstColumnIdx = clamp(
    mb.firstColumnIdx + chars,
    0,
    max(0, len(mb.content) - mb.dims.width))
  render(mb)

proc resetScroll*(
    mb: var MessageBuffer[ScrollText]) =
  mb.firstColumnIdx = 0
  mb.firstLineIdx = 0
  render(mb)
